# Copyright 2021 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import deprecation

from tobiko import version


def deprecated(deprecated_in: str = None,
               removed_in: str = None,
               details=""):
    return deprecation.deprecated(deprecated_in=deprecated_in,
                                  removed_in=removed_in,
                                  current_version=version.version,
                                  details=details)
