# Copyright (c) 2022 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

import tobiko
from tobiko.openstack import designate
from tobiko.openstack import heat
from tobiko.openstack.stacks import _hot


class DesignateZoneStackFixture(heat.HeatStackFixture):
    template = _hot.heat_template_file('designate/zone.yaml')

    @property
    def zone_name(self) -> str:
        return tobiko.get_fixture_name(self).lower() + '.'

    @property
    def zone_details(self) -> typing.Mapping[str, typing.Any]:
        return designate.get_designate_zone(self.zone_id)
