# Copyright (c) 2020 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import tobiko
from tobiko.openstack import keystone
from tobiko.openstack import stacks
from tobiko.tests.functional.openstack.stacks import test_centos


@keystone.skip_unless_has_keystone_credentials()
class RedHatServerStackTest(test_centos.CentosServerStackTest):
    """Test Red Hat server instance"""

    #: Stack of resources with a server attached to a floating IP
    stack = tobiko.required_fixture(stacks.RedHatServerStackFixture)
